import os
import sys
import time
import math
import torch
import shutil
import random
import logging
import argparse
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt

import torch.nn as nn
from torch import optim
import torch.backends.cudnn as cudnn
from torch.utils.data import DataLoader, random_split

from nets.unet import UNet
from utils.metrics import dice as dice_all
from utils.metrics import batch_dice,compute_dice
from utils.losses import BinaryDiceLoss
from utils.util import set_logging,Logger,read_list,plot_base,plot_dice,AverageMeter,save_img_inter,plot_dice_loss
from dataloader.dataset import BaseDataSets,PatientBatchSampler

def train_net(start_time,base_dir,data_path,train_list,val_list,device,img_mode='',
            lr_scheduler='warmupMultistep',
            max_epoch=81,
            batch_size=80,
            images_rate=0.1,
            base_lr=0.001,
            weight_decay=0.0003,
            optim_name='adam',
            loss_name='bce',
            ):
    step_num_lr = 4
    local_vars_dict = {}
    for var in train_net.__code__.co_varnames: 
        if var == 'local_vars_dict':
            break
        local_vars_dict[var] = locals()[var]

    warm_up_epochs = int(max_epoch * 0.1)
    
    image_channels = 1  # image_channels=3 for RGB images
    class_num = 1       # For 1 class and background, use class_num=1
    if class_num==1:
        mask_name='masks'
    elif class_num==3:
        mask_name='masks_all'

    """network"""
    net = UNet(image_channels,class_num,32,bilinear=False)
    net.to(device=device)
    net_name = str(net)[0:str(net).find('(')]

    train_dataset = BaseDataSets(data_path,"train",img_mode,mask_name,train_list,images_rate)
    val_dataset = BaseDataSets(data_path,"val",img_mode,mask_name,val_list)
    val_dataset_3d = BaseDataSets(data_path,"val_3d",img_mode,mask_name,val_list_full)

    n_train = train_dataset.__len__()
    n_val = val_dataset.__len__()
    def worker_init_fn(worker_id):
        random.seed(1111 + worker_id)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, num_workers=16, pin_memory=True, drop_last=True,worker_init_fn=worker_init_fn)
    # train_loader = DataLoader(train_dataset,batch_sampler=batch_sampler, num_workers=16, pin_memory=True,worker_init_fn=worker_init_fn)
    val_loader_2d = DataLoader(val_dataset, batch_size=batch_size, shuffle=False, num_workers=16, pin_memory=True, drop_last=False,worker_init_fn=worker_init_fn)
    slices_list = val_dataset_3d.__sampleList__()
    batch_samplerPatient = PatientBatchSampler(slices_list,patientID_list)
    val_loader_3d = DataLoader(val_dataset_3d,batch_sampler=batch_samplerPatient, num_workers=16, pin_memory=True,worker_init_fn=worker_init_fn)
    
    day_time = start_time.split(' ')
    time_str = str(day_time[0].split('-')[1] + day_time[0].split('-')[2] + day_time[1].split(':')[0] + day_time[1].split(':')[1])

    if optim_name=='adam':
        optimizer = optim.Adam(net.parameters(), lr=base_lr, weight_decay=weight_decay)
    elif optim_name=='sgd':
        optimizer = optim.SGD(net.parameters(), lr=base_lr, momentum=0.9,weight_decay=weight_decay)
    elif optim_name=='adamW':
        optimizer = optim.AdamW(net.parameters(), lr=base_lr, weight_decay=weight_decay)    
        
    if lr_scheduler=='warmupMultistep':
        # warm_up_with_multistep_lr
        if step_num_lr == 2:
            lr1,lr2 = int(max_epoch*0.3) ,int(max_epoch*0.6)
            lr_milestones = [lr1,lr2]
        elif step_num_lr == 3:    
            lr1,lr2,lr3 = int(max_epoch*0.25) , int(max_epoch*0.4) , int(max_epoch*0.6)
            lr_milestones = [lr1,lr2,lr3]
        elif step_num_lr == 4:    
            lr1,lr2,lr3,lr4 = int(max_epoch*0.15) , int(max_epoch*0.35) , int(max_epoch*0.55) , int(max_epoch*0.7)
            lr_milestones = [lr1,lr2,lr3,lr4]
        warm_up_with_multistep_lr = lambda epoch: (epoch+1) / warm_up_epochs if epoch < warm_up_epochs \
                                                else 0.1**len([m for m in lr_milestones if m <= epoch])
        scheduler_lr = optim.lr_scheduler.LambdaLR(optimizer,lr_lambda = warm_up_with_multistep_lr)
    elif lr_scheduler=='warmupCosine':
        # warm_up_with_cosine_lr
        warm_up_with_cosine_lr = lambda epoch: (epoch+1) / warm_up_epochs if epoch < warm_up_epochs \
                                else 0.5 * ( math.cos((epoch - warm_up_epochs) /(max_epoch - warm_up_epochs) * math.pi) + 1)
        scheduler_lr = optim.lr_scheduler.LambdaLR(optimizer,lr_lambda = warm_up_with_cosine_lr)
    elif lr_scheduler=='autoReduce':
        scheduler_lr = optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min',factor=0.5, patience=6, verbose=True, cooldown=2,min_lr=0)
    if loss_name=='bce':
        criterion = nn.BCELoss()

    dice_loss = BinaryDiceLoss()
    optimizer_name = str(optimizer)[0:str(optimizer).find('(')]

    param_str = "Starting training:"
    for var in list(local_vars_dict.keys()):
        if var != 'device':
            var_value = local_vars_dict[var]
            param_str += "\n\t" + var + ":" + " "*(15-len(var)) + str(var_value)
    
    logging.info(param_str+f'''\n\tNet Name:\t\t{net_name}\n\tInput Channel:\t{image_channels}\n\tClasses Num:\t{class_num}\n\tImages Shape:\t{img_h}*{img_w}''')

    train_log = AverageMeter()
    val_log = AverageMeter()
    lr_curve = list()
    for epoch in range(max_epoch):
        # train the model
        net.train()
        with tqdm(total=n_train, desc=f'Epoch {epoch + 1}/{max_epoch}', unit='img', leave=is_leave) as pbar:
            for i,batch in enumerate(train_loader):
                # forward
                imgs = batch['image']
                true_masks = batch['mask']
                slice_name = batch['idx']
                # print("batch idxs:",batch['idx'])
                assert imgs.shape[1] == image_channels, \
                    f'Network has been defined with {image_channels} input channels, ' \
                    f'but loaded images have {imgs.shape[1]} channels. Please check that ' \
                    'the images are loaded correctly.'

                imgs = imgs.to(device=device, dtype=torch.float32)
                true_masks = true_masks.to(device=device, dtype=torch.float32)
                masks_pred = net(imgs)
                #for ce loss and dice loss
                loss_base = criterion(masks_pred, true_masks)
                loss_dice = dice_loss(masks_pred, true_masks)
                loss = loss_base * 10 + loss_dice * 7
                train_log.add_value({"loss": loss.item()}, n=1)  

                '''backward'''
                loss.backward()
                '''update weights'''
                optimizer.step()        # update parameters of net
                optimizer.zero_grad()   # reset gradient


                pred = (masks_pred > 0.5).float()
                dice_sum,num = batch_dice(pred.cpu().data, true_masks.cpu())
                dice = dice_sum / num

                train_log.add_value({"dice": dice,}, n=1)
                if set_args == False:
                    pbar.update(batch_size)
                
            train_log.updata_avg()
            mean_loss = train_log.res_dict["loss"][epoch]
            mean_dice = train_log.res_dict["dice"][epoch]

       # validate the model
        net.eval()
        n_val_2d = len(val_loader_2d)  
        n_val_3d = len(val_loader_3d)  
        if epoch % val_3d_interval==0:
            compute_3d = True
            n_val = n_val_3d
            val_loader = val_loader_3d
        else:
            compute_3d = False
            n_val = n_val_2d
            val_loader = val_loader_2d

        with tqdm(total=n_val, desc='Validation round', unit='batch', leave=False) as pbar:
            for j,batch_val in enumerate(val_loader):
                imgs = batch_val['image']
                true_masks = batch_val['mask']

                imgs = imgs.to(device=device, dtype=torch.float32)
                true_masks = true_masks.to(device=device, dtype=torch.float32)
                if compute_3d==True:
                    batch_num = imgs.shape[0]
                    batch_fore = int(batch_num / 3)
                    imgs_fore = imgs[:batch_fore]
                    imgs_mid = imgs[batch_fore:batch_fore*2]
                    imgs_after = imgs[batch_fore*2:]

                    with torch.no_grad():
                        mask_pred_fore = net(imgs_fore)
                        mask_pred_mid = net(imgs_mid)
                        mask_pred_after = net(imgs_after)
                    mask_pred = torch.cat([mask_pred_fore, mask_pred_mid, mask_pred_after], dim=0)
                    # dice
                    pred = mask_pred.ge(0.5).float()
                    pred_np = pred.cpu().data.numpy().astype("uint8")
                    true_np = true_masks.cpu().numpy().astype("uint8")
                    if class_num == 1:
                        # 3D dice for class number == 1
                        dice_val_3d = dice_all(pred_np, true_np)
                    elif class_num == 3:
                        # 3D dice for class number == 3
                        dice_val_3d_1 = dice_all(pred_np[:,0,:,:], true_np[:,0,:,:])
                        dice_val_3d_2 = dice_all(pred_np[:,1,:,:], true_np[:,1,:,:])
                        dice_val_3d_3 = dice_all(pred_np[:,2,:,:], true_np[:,2,:,:])
                        dice_val_3d = (dice_val_3d_1 + dice_val_3d_2 + dice_val_3d_3) / 3
                    val_log.add_value({"dice_3d": dice_val_3d}, n=1)
                    dice_val_sum,nidus_num,nidus_start = compute_dice(pred.cpu().data, true_masks.cpu(),deNoNidus=True)
                    nidus_end = nidus_start+nidus_num-1
                    mask_pred = mask_pred[nidus_start:nidus_end+1]
                    true_masks = true_masks[nidus_start:nidus_end+1]
                else:
                    with torch.no_grad():
                        mask_pred = net(imgs)
                    # 2D dice
                    pred = mask_pred.ge(0.5).float()
                    dice_val_sum,nidus_num = batch_dice(pred.cpu().data, true_masks.cpu())
                val_log.add_value({"dice": dice_val_sum}, n=nidus_num)
                
                # vlaid loss
                loss_val_base = criterion(mask_pred, true_masks)
                loss_val_dice = dice_loss(mask_pred, true_masks)
                loss_val = loss_val_base * 10.0 + loss_val_dice * 7.0
                val_log.add_value({"loss": loss_val.item()}, n=1)
                if set_args == False:
                    pbar.update()

            val_log.updata_avg()
            valid_loss_mean = val_log.res_dict["loss"][epoch]
            valid_dice_mean = val_log.res_dict["dice"][epoch]

            if epoch % val_3d_interval==0:
                valid_dice_3d_mean = val_log.res_dict["dice_3d"][-1]
                logging.info(
                    'Epoch:[{:0>3}/{:0>3}], Train Loss: {:.4f} , Val Loss: {:.4f} ,Train Dice: 2d {:.4f},  Val Dice: 2d  {:.4f} 3D {:.4f}'.format(
                            epoch,max_epoch,          mean_loss,  valid_loss_mean,            mean_dice ,       valid_dice_mean , valid_dice_3d_mean))
            else:
                logging.info(
                    'Epoch:[{:0>3}/{:0>3}], Train Loss: {:.4f} , Val Loss: {:.4f} ,Train Dice: 2d {:.4f},  Val Dice: 2d  {:.4f}'.format(
                            epoch,max_epoch,          mean_loss,  valid_loss_mean,            mean_dice ,       valid_dice_mean))
        
        ## update lr
        if lr_scheduler=='autoReduce':
            scheduler_lr.step(valid_loss_mean)
        else:
            scheduler_lr.step()
        lr_epoch = optimizer.param_groups[0]['lr']
        lr_curve.append(lr_epoch)

    return train_log.res_dict,val_log.res_dict,lr_curve,net

def main():
    base_dir = 'res-HECK-full-train'
    img_mode = 'ct'  # ct  pet  t2  t1ce 
 
    if set_args==True:
        args = set_argparse()
        print('WARNING!!! Using argparse for parameters to obtain ')
        base_dir = args.base_dir
        img_mode = args.img_mode
    assert 'res-' in base_dir, \
            f'base_dir should include string:\'res-\',but base_dir is \'{base_dir}\'.'
    base_dir = base_dir.replace('res-',f'res-{img_mode}-',1)
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)
    backup_code(base_dir)
    log_path = os.path.join(base_dir, 'training.log') 
    sys.stdout = Logger(log_path=log_path)
    set_logging(log_path=log_path)
    set_random_seed(seed_num=1111)

    """GPU ID"""
    gpu_list = [0] #[0,1]
    gpu_list_str = ','.join(map(str, gpu_list))
    os.environ.setdefault("CUDA_VISIBLE_DEVICES", gpu_list_str)
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    logging.info(f'Using device : {device}\n'
                 f'\tGPU ID is [{os.environ["CUDA_VISIBLE_DEVICES"]}],using {torch.cuda.device_count()} device\n'
                 f'\tdevice name:{torch.cuda.get_device_name(0)}')

    start_time = time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time()))
    
    global img_h,img_w
    if img_mode=='ct' or img_mode=='pet':
        data_path = "../data/Hecktor_slice" 
        img_h,img_w = 144,144
    else:
        data_path = "../data/BraTS_slice" 
        img_h,img_w = 160,160
    train_list = 'randP1_slice_nidus_train.list'
    val_list = 'randP1_slice_nidus_val.list'
    global val_list_full
    val_list_full = 'randP1_slice_all_val.list'
    global patientID_list
    patientID_list = read_list(data_path+'/randP1_volume_val.list')
    time_tic = time.time()
    global is_leave
    global val_3d_interval
    val_3d_interval = 4 
    logging.info('============ Start  train ==============')
    if set_args==True:
        ########## Using Argument Parser ##########
        is_leave = False
        train_log,val_log,lr_curve,net = train_net(start_time,base_dir,data_path,args.train_list,args.val_list,device,
            args.img_mode,args.lr_scheduler,args.max_epoch,args.batch_size,args.images_rate,args.base_lr,args.weight_decay,args.optim_name,args.loss_name)
    else:
        is_leave = True
        train_log,val_log,lr_curve,net = train_net(start_time,base_dir,data_path,train_list,val_list,device,img_mode)

    net_name = str(net)[0:str(net).find('(')]
    mode_path_name = base_dir + '/' + f'model_{net_name}_last.pth'
    torch.save(net, mode_path_name)
    logging.info('Model saved !')    

    """Plot"""
    plot_dice_loss(train_log,val_log,val_3d_interval,lr_curve,base_dir)

    time_toc = time.time()
    time_s = time_toc - time_tic
    time_end = time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time_toc))
    logging.info("Train finished time: {}".format(time_end))
    logging.info("Time consuming : {:.2f} min in train and test".format(time_s / 60))

def set_argparse():
    parser = argparse.ArgumentParser()
    parser.add_argument('--base_dir', type=str,default='res-HECK-train-full', 
                        help='base dir name')
    parser.add_argument('--train_list', type=str,default='randP1_slice_nidus_train.list', 
                        help='a list of train data')
    parser.add_argument('--val_list', type=str,default='randP1_slice_nidus_val.list', 
                        help='a list of val data')
    parser.add_argument('--img_mode', type=str,default='ct', # ct pet  t2 t1ce t1 flair
                        help='medical images mode')
    parser.add_argument('--max_epoch', type=int,default=81, 
                        help='maximum epoch')
    parser.add_argument('--batch_size', type=int,default=80,
                        help='batch size per gpu')
    parser.add_argument('--images_rate', type=float,default=0.1,
                        help='images rate')
    parser.add_argument('--base_lr', type=float,default=0.01,
                        help='segmentation network learning rate')
    parser.add_argument('--weight_decay', type=float,default=1e-3,
                        help='weight decay(L2 Regularization)')   
    parser.add_argument('--optim_name', type=str,default='adam', 
                        help='optimizer name')
    parser.add_argument('--loss_name', type=str,default='bce', 
                        help='loss name')   
    parser.add_argument('--lr_scheduler', type=str,default='warmupMultistep', 
                        help='lr scheduler')      

    args = parser.parse_args()
    return args

def set_random_seed(seed_num):
    if seed_num != '':
        logging.info(f'set random seed: {seed_num}')
        cudnn.benchmark = False    
        cudnn.deterministic = True  
        random.seed(seed_num)    
        np.random.seed(seed_num) 
        torch.manual_seed(seed_num)
        torch.cuda.manual_seed(seed_num)

def backup_code(base_dir):
    code_path = os.path.join(base_dir, 'code') 
    if not os.path.exists(code_path):
        os.makedirs(code_path)
    train_name = os.path.basename(__file__)
    dataset_name = 'dataset.py'
    shutil.copy('dataloader/' + dataset_name, code_path + '/' + dataset_name)
    shutil.copy(train_name, code_path + '/' + train_name)

if __name__ == '__main__':
    set_args = True
    main()
